# Silvius microphone client based on Tanel's client.py
__author__ = "dwk"

import argparse
from ws4py.client.threadedclient import WebSocketClient
import threading
import sys
import urllib.request, urllib.parse, urllib.error
import json

reconnect_mode = False
fatal_error = False


class MyClient(WebSocketClient):
    def __init__(
        self,
        url,
        mic=1,
        protocols=None,
        extensions=None,
        heartbeat_freq=None,
        byterate=16000,
        show_hypotheses=True,
        save_adaptation_state_filename=None,
        send_adaptation_state_filename=None,
        audio_gate=0,
    ):
        super(MyClient, self).__init__(url, protocols, extensions, heartbeat_freq)
        self.mic = mic
        self.show_hypotheses = show_hypotheses
        self.byterate = byterate
        self.save_adaptation_state_filename = save_adaptation_state_filename
        self.send_adaptation_state_filename = send_adaptation_state_filename
        self.chunk = 0
        self.audio_gate = audio_gate

    def send_data(self, data):
        self.send(data, binary=True)

    def opened(self):
        import pyaudio
        import audioop

        pa = pyaudio.PyAudio()
        sample_rate = self.byterate
        stream = None

        while stream is None:
            try:
                # try adjusting this if you want fewer network packets
                self.chunk = 2048 * 2 * sample_rate // self.byterate

                mic = self.mic
                if mic == -1:
                    mic = pa.get_default_input_device_info()["index"]
                    print("Selecting default mic", file=sys.stderr)
                print("Using mic #", mic, file=sys.stderr)
                stream = pa.open(
                    rate=sample_rate,
                    format=pyaudio.paInt16,
                    channels=1,
                    input=True,
                    input_device_index=mic,
                    frames_per_buffer=self.chunk,
                )
            except IOError as e:
                if e.errno == -9997 or e.errno == "Invalid sample rate":
                    new_sample_rate = int(
                        pa.get_device_info_by_index(mic)["defaultSampleRate"]
                    )
                    if sample_rate != new_sample_rate:
                        sample_rate = new_sample_rate
                        continue
                print("\n", e, file=sys.stderr)
                print(
                    "\nCould not open microphone. Please try a different device.",
                    file=sys.stderr,
                )
                global fatal_error
                fatal_error = True
                sys.exit(0)

        def mic_to_ws():  # uses stream
            try:
                print("\nLISTENING TO MICROPHONE", file=sys.stderr)
                last_state = None
                while True:
                    data = stream.read(self.chunk)
                    if self.audio_gate > 0:
                        rms = audioop.rms(data, 2)
                        if rms < self.audio_gate:
                            data = "\00" * len(data)
                    # if sample_chan == 2:
                    #    data = audioop.tomono(data, 2, 1, 1)
                    if sample_rate != self.byterate:
                        (data, last_state) = audioop.ratecv(
                            data, 2, 1, sample_rate, self.byterate, last_state
                        )

                    self.send_data(data)
            except IOError as e:
                # usually a broken pipe
                print(e)
            except AttributeError:
                # currently raised when the socket gets closed by main thread
                pass

            # to voluntarily close the connection, we would use
            # self.send_data("")
            # self.send("EOS")

            try:
                self.close()
            except IOError:
                pass

        threading.Thread(target=mic_to_ws).start()

    def received_message(self, m):

        response = json.loads(str(m))
        # print >> sys.stderr, "RESPONSE:", response
        # print >> sys.stderr, "JSON was:", m
        if response["status"] == 0:
            if "result" in response:
                trans = response["result"]["hypotheses"][0]["transcript"]
                if response["result"]["final"]:
                    if self.show_hypotheses:
                        print("\r%s" % trans.replace("\n", "\\n"), file=sys.stderr)
                    # print("%s" % trans.replace("\n", "\\n"), flush=True)  # final result!
                    print(m, flush=True)
                elif self.show_hypotheses:
                    print_trans = trans.replace("\n", "\\n")
                    if len(print_trans) > 80:
                        print_trans = "... %s" % print_trans[-76:]
                    print("\r%s" % print_trans, end=" ", file=sys.stderr)
            if "adaptation_state" in response:
                if self.save_adaptation_state_filename:
                    print(
                        "Saving adaptation state to %s"
                        % self.save_adaptation_state_filename,
                        file=sys.stderr,
                    )
                    with open(self.save_adaptation_state_filename, "w") as f:
                        f.write(json.dumps(response["adaptation_state"]))
        else:
            print(
                "Received error from server (status %d)" % response["status"],
                file=sys.stderr,
            )
            if "message" in response:
                print("Error message:", response["message"], file=sys.stderr)

            global reconnect_mode
            if reconnect_mode:
                import time

                print("Sleeping for five seconds before reconnecting", file=sys.stderr)
                time.sleep(5)

    def closed(self, code, reason=None):
        # print "Websocket closed() called"
        # print >> sys.stderr
        pass


def setup():
    content_type = "audio/x-raw, layout=(string)interleaved, rate=(int)16000, format=(string)S16LE, channels=(int)1"
    path = "client/ws/speech"

    parser = argparse.ArgumentParser(description="Microphone client for silvius")
    parser.add_argument(
        "-s",
        "--server",
        default="localhost",
        dest="server",
        help="Speech-recognition server",
    )
    parser.add_argument("-p", "--port", default="8019", dest="port", help="Server port")
    # parser.add_argument('-r', '--rate', default=16000, dest="rate", type=int, help="Rate in bytes/sec at which audio should be sent to the server.")
    parser.add_argument(
        "-d",
        "--device",
        default="-1",
        dest="device",
        type=int,
        help="Select a different microphone (give device ID)",
    )
    parser.add_argument(
        "-k",
        "--keep-going",
        action="store_true",
        help="Keep reconnecting to the server after periods of silence",
    )
    parser.add_argument("--save-adaptation-state", help="Save adaptation state to file")
    parser.add_argument(
        "--send-adaptation-state", help="Send adaptation state from file"
    )
    parser.add_argument(
        "--content-type",
        default=content_type,
        help="Use the specified content type (default is " + content_type + ")",
    )
    parser.add_argument(
        "--hypotheses",
        default=True,
        type=int,
        help="Show partial recognition hypotheses (default: 1)",
    )
    parser.add_argument(
        "-g",
        "--audio-gate",
        default=0,
        type=int,
        help="Audio-gate level to reduce detections when not talking",
    )
    args = parser.parse_args()

    content_type = args.content_type
    print("Content-Type:", content_type, file=sys.stderr)

    if args.keep_going:
        global reconnect_mode
        global fatal_error
        reconnect_mode = True
        while fatal_error == False:
            print("Reconnecting...", file=sys.stderr)
            run(args, content_type, path)
    else:
        run(args, content_type, path)


def run(args, content_type, path):
    uri = "ws://%s:%s/%s?%s" % (
        args.server,
        args.port,
        path,
        urllib.parse.urlencode([("content-type", content_type)]),
    )
    print("Connecting to", uri, file=sys.stderr)

    ws = MyClient(
        uri,
        byterate=16000,
        mic=args.device,
        show_hypotheses=args.hypotheses,
        save_adaptation_state_filename=args.save_adaptation_state,
        send_adaptation_state_filename=args.send_adaptation_state,
        audio_gate=args.audio_gate,
    )
    ws.connect()
    # result = ws.get_full_hyp()
    # print result.encode('utf-8')
    ws.run_forever()


def main():
    try:
        setup()
    except KeyboardInterrupt:
        print("\nexiting...", file=sys.stderr)


if __name__ == "__main__":
    main()



import time
import os
import subprocess
import sys

import attr
import pytest


@attr.dataclass
class VirtualDisplay:
    name: str
    process: subprocess.Popen


@pytest.fixture(name="virtual_display", scope="session")
def _virtual_display():

    original_display = os.environ.get("DISPLAY", None)
    name = ":6"
    os.environ["DISPLAY"] = name

    proc = subprocess.Popen(
        ["/usr/bin/Xvfb", name, "-screen", "0", "1920x1080x24+32", "-fbdir", "/var/tmp"]
    )
    virtual_display = VirtualDisplay(name, proc)
    time.sleep(1.0)
    try:
        yield virtual_display
    finally:
        proc.terminate()

        if original_display is None:
            del os.environ["DISPLAY"]
        else:
            os.environ["DISPLAY"] = original_display


@pytest.fixture(name="window_manager")
def _window_manager(virtual_display):
    proc = subprocess.Popen(["/usr/bin/i3"])
    time.sleep(1)
    try:
        yield
    finally:
        proc.terminate()


@pytest.fixture(name="turtle_window")
def _turtle_window(window_manager):
    turtle = subprocess.Popen(["/usr/bin/python3", "-m", "turtle"])
    time.sleep(2)
    try:
        yield
    finally:
        turtle.terminate()


@pytest.fixture(name="idle_window")
def _idle_window(window_manager):
    idle = subprocess.Popen(["/usr/bin/idle"])
    time.sleep(1)
    try:
        yield
    finally:
        idle.terminate()


@pytest.fixture(name="hash_seed")
def _hash_seed():
    original_hash_seed = os.environ.get("PYTHONHASHSEED")
    os.environ["PYTHONHASHSEED"] = "123"
    try:
        yield
    finally:
        if original_hash_seed is None:
            del os.environ["PYTHONHASHSEED"]
        else:
            os.environ["PYTHONHASHSEED"] = original_hash_seed


@pytest.fixture(name="config_dir", autouse=True)
def _config_dir(tmp_path):
    os.environ["VOCA_CONFIG_DIR"] = str(tmp_path)
    return tmp_path



import platform

from voca import platforms


@platforms.implementation(platforms.System.DARWIN)
def get_platform():
    return "This is Darwin"


@platforms.implementation(platforms.System.LINUX)
def get_platform():
    return "This is Linux"


@platforms.implementation(platforms.System.WINDOWS)
def get_platform():
    return "This is Windows"


def test_implementation():
    assert get_platform() == "This is " + platform.system()


def test_dict_accessible():
    assert len(set(get_platform.implementations)) == 3



import contextlib
import os
import json
import secrets
import subprocess
import string
import textwrap

from click.testing import CliRunner
import pytest

from voca import cli

from tests import helpers
from tests.helpers import make_command


def test_main():
    runner = CliRunner()
    result = runner.invoke(cli.cli, [])

    assert result.output.startswith("Usage")
    assert result.exit_code == 0


@pytest.mark.usefixtures("virtual_display")
def test_strict():
    """Strict mode, not eager mode, executes final commands."""
    utterances = ["say alpha", "say bravo", "mode", "say charlie", "say delta"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage", "-i", "voca.plugins.basic"], input=lines)

    expected = ["KEY_A", "KEY_B"]
    assert typed == expected


@pytest.mark.usefixtures("virtual_display")
def test_eager():
    """Eager mode, not strict mode, executes non-final commands."""
    utterances = ["say alpha", "say bravo", "mode", "say charlie", "say delta"]

    rows = [make_command(utterance, final=False) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage", "-i", "voca.plugins.basic"], input=lines)

    expected = ["KEY_C", "KEY_D"]
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "turtle_window")
def test_context_always():

    utterances = ["hit alpha", "hit bravo"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            [
                "manage",
                "-i",
                "voca.plugins.basic",
                "-i",
                "voca.plugins.yes",
                "-i",
                "voca.plugins.no",
            ],
            input=lines,
        )

    expected = ["KEY_D", "KEY_E", "KEY_F"] * 2
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "turtle_window")
def test_context_never():

    utterances = ["nope", "nope"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage"], input=lines)

    expected = []
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "turtle_window")
def test_app_context_matches():
    """The spec is matched and the action is executed if the app is open."""

    utterances = ["check", "check"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            ["manage", "-i", "voca.plugins.basic", "-i", "voca.plugins.turtle_context"],
            input=lines,
        )

    expected = ["KEY_T", "KEY_U", "KEY_R", "KEY_T"] * 2
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "idle_window")
def test_app_context_does_not_match():
    """The spec is matched and the action is executed if the app is open."""

    utterances = ["check", "check"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage"], input=lines)

    expected = []
    assert typed == expected


@pytest.mark.usefixtures("virtual_display")
def test_config_module(tmp_path):
    """Can add grammars from user config modules."""

    # Given
    source = textwrap.dedent(
        """\
        from voca import utils
        from voca.plugins import basic


        registry = utils.Registry()
        wrapper = utils.Wrapper(registry)


        @registry.register('"my config"')
        async def _type_xy(_):
            await basic.write("xy")
        """
    )

    user_modules_path = tmp_path / "user_modules"
    user_modules_path.mkdir()
    (user_modules_path / "my_module.py").write_text(source)

    # When
    utterances = ["my config"]
    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage", "-i", "user_modules.my_module"], input=lines)

    # Then
    expected = ["KEY_X", "KEY_Y"]
    assert typed == expected



"""The manager starts worker processes and delegates input commands to workers.

The manager is the primary supervisor process, restarting its workers to allow
reloading and error handling.

"""

from __future__ import annotations

import functools
import os
import itertools
import sys
import pathlib
import subprocess
import json
import contextlib
import copy

from typing import List
from typing import Optional
from typing import Dict
from typing import Set

import attr
import trio
import eliot


from voca import plugins
from voca import utils
from voca import streaming
from voca import log


def handle_unexpected_worker_bytes(message: bytes):
    eliot.Message.log(message_type="unexpected_worker_output", message=message)


def worker_cli(should_log, module_names: Optional[List[str]] = None) -> List[str]:
    """Build the list of strings for invoking a worker subprocess."""
    if module_names is None:
        # TODO This case should be handled in the worker, not the manager.
        module_names = utils.get_module_names()

    log_arg = "--log" if should_log else "--no-log"
    prefix = [sys.executable, "-m", "voca", log_arg, "worker"]
    command = prefix.copy()
    for module_name in module_names:
        command += ["-i", module_name]
    return command


@log.log_async_call
async def replay_child_messages(child: trio.Process) -> None:
    """Log the child's messages to the parent's stdout."""
    async for message_from_child in streaming.TerminatedFrameReceiver(
        child.stdout, b"\n"
    ):
        message_str = message_from_child.decode()
        try:
            message_dict = json.loads(message_str)
        except json.decoder.JSONDecodeError:
            handle_unexpected_worker_bytes(message_from_child)
        else:
            print(message_dict)


@log.log_call
def set_state(data: Dict[str, dict], state: Dict[str, dict]):
    """Handle switching between ``eager`` and ``strict`` mode."""
    # XXX maybe this should be in the worker.
    state = copy.deepcopy(state)
    body = data["result"]["hypotheses"][0]["transcript"]
    command, _space, _args = body.partition(" ")
    if command == "mode":
        state["modes"]["strict"] = not state["modes"]["strict"]
        return state

    return None


@log.log_async_call
async def delegate_task(
    data: Dict, worker: trio.Process, state: dict, action: eliot.Action
):
    """Send input data to worker process over std streams."""

    wrapped_data = dict(
        **data, state=state, eliot_task_id=action.serialize_task_id().decode()
    )
    await worker.stdin.send_all(json.dumps(wrapped_data).encode() + b"\n")


@attr.s
class Pool:
    num_workers: int = attr.ib(default=1)
    should_log: bool = attr.ib(default=True)
    module_names: List[str] = attr.ib(factory=list)
    processes: Set[trio.Process] = attr.ib(factory=set)

    def start(self) -> None:
        """Start a new process."""
        for _ in range(self.num_workers):
            self.add_new_process()

    def get_process(self) -> trio.Process:
        """Pop an process outof the pool and return it."""
        return self.processes.pop()

    def add_new_process(self) -> None:
        """Start a new process and add it to the pool."""
        self.processes.add(
            trio.Process(
                worker_cli(self.should_log, self.module_names),
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
        )


@log.log_async_call
async def run_worker(data: dict, state: dict, pool: Pool):
    """Get a process from the pool, send a job to it. Replace that worker when it quits."""
    with eliot.start_action(action_type="run_with_work") as action:
        worker = pool.get_process()

        await delegate_task(data=data, state=state, worker=worker, action=action)
        await replay_child_messages(worker)

        await worker.wait()
        pool.add_new_process()


@log.log_async_call
async def process_stream(
    receiver, num_workers: int, should_log: bool, module_names: Optional[List[str]]
):
    """Handle all the commands coming in by delegating them to workers."""

    state = {"modes": {"strict": True}}

    pool = Pool(num_workers, should_log=should_log, module_names=module_names)
    pool.start()

    async for message_bytes in receiver:
        message = message_bytes.decode()

        with eliot.start_action(state=state):
            try:
                data = json.loads(message)
            except json.JSONDecodeError:
                handle_unexpected_worker_bytes(message_bytes)
                continue
            if "result" not in data.keys():
                # Received a log, not a command.
                print(message)
                continue
            # Handle state changes.
            # XXX maybe this should be in the worker, but where to store state?
            maybe_new_state = set_state(data, state)
            if maybe_new_state is not None:
                state = maybe_new_state
                continue

            # This logic could be moved into worker/plugin to allow for more modes.
            if not data["result"]["final"] and state["modes"]["strict"]:
                continue
            if data["result"]["final"] and not state["modes"]["strict"]:
                continue

            await run_worker(data=data, state=state, pool=pool)


@log.log_async_call
async def async_main(should_log, module_names: Optional[List[str]], num_workers: int):
    """Read newline-separated inputs on stdin, and process them."""

    stream = trio._unix_pipes.PipeReceiveStream(os.dup(0))
    receiver = streaming.TerminatedFrameReceiver(stream, b"\n")

    await process_stream(
        receiver,
        num_workers=num_workers,
        should_log=should_log,
        module_names=module_names,
    )


@utils.public
@log.log_call
def main(should_log: bool, module_names: Optional[List[str]], num_workers: int):
    """Start the event loop."""
    trio.run(functools.partial(async_main, should_log, module_names, num_workers))



"""Define a meta-path finder and loader for importing modules from a synthetic
namespace.
"""

from __future__ import annotations

import contextlib
import collections
import sys
import types
import importlib
import importlib.machinery
import importlib.util

from typing import Generator
from typing import ContextManager
from typing import Optional
from typing import Dict
from typing import List
from typing import Any
from typing import DefaultDict
from typing import TypeVar


import attr

from voca import utils

T = TypeVar("T")


@attr.s(auto_attribs=True)
class PathLoader:
    namespace: dict
    fullname: str
    path: Optional[str]
    target: types.ModuleType

    def create_module(self, spec: importlib.machinery.ModuleSpec) -> types.ModuleType:
        """Create a module with the name and namespace."""
        module = types.ModuleType(spec.name)
        module.__dict__.update(self.namespace)

        return module

    def exec_module(self, module):
        """No-op to match the Loader interface."""
        pass


@contextlib.contextmanager
def skipping_finder(finder: importlib.abc.MetaPathFinder) -> Generator:
    """Context manager for importing modules while excluding the ``finder`` from ``sys.meta_path``."""

    def placeholder_find_spec(_fullname, _path, _target):
        return None

    finder_placeholder = types.SimpleNamespace(find_spec=placeholder_find_spec)

    sys.meta_path[sys.meta_path.index(finder)] = finder_placeholder
    try:
        yield
    finally:
        sys.meta_path[sys.meta_path.index(finder_placeholder)] = finder


@contextlib.contextmanager
def skipping_module_in_sys_modules(module_name: star) -> Generator:
    """Context manager for importing modules while skipping any cached entry in ``sys.modules``."""
    # Use a sentinel to check if we need to restore parent module to sys.modules.
    sentinel = object()
    parent_module = sys.modules.pop(module_name, sentinel)
    try:
        yield
    finally:
        # Restore parent if necessary.
        if parent_module is not sentinel:

            # XXX Not sure if this is the right choice, or if it should leave
            # the existing entry in place.

            # Restore original module to sys.modules, bumping any entry created
            # during the `yield` above.
            sys.modules[module_name] = parent_module


@contextlib.contextmanager
def finder_patch(finder: importlib.abc.MetaPathFinder) -> Generator:
    """Context manager for importing modules with ``finder`` on ``sys.meta_path``."""
    sys.meta_path.insert(0, finder)
    try:
        yield
    finally:
        sys.meta_path.remove(finder)


@attr.s(auto_attribs=True)
class PathFinder:
    modules_to_handle: List[str]
    fullname_to_vars: Dict[str, Dict[str, Any]]

    def find_spec(
        self, fullname: str, path: Optional[str], target=Optional[types.ModuleType]
    ) -> Spec:
        """Build a spec for modules in ``self.modules_to_handle``, behave normally otherwise."""
        if fullname not in self.modules_to_handle:
            # Return a spec using the default behavior ignoring sys.modules and
            # this finder.
            parent_name = ".".join(fullname.split(".")[:-1])

            with skipping_finder(self), skipping_module_in_sys_modules(parent_name):

                return importlib.util.find_spec(fullname, path)

        variables = self.fullname_to_vars.get(fullname, {})
        loader = PathLoader(variables, fullname, path, target)

        return Spec(name=fullname, loader=loader)


@attr.s
class Spec:
    name = attr.ib()
    loader = attr.ib()
    submodule_search_locations = attr.ib(factory=list)
    has_location = attr.ib(default=True)
    cached = attr.ib(default=False)
    origin = attr.ib(default=None)


def ancestors(items: List[T]) -> Generator[T, List[T], None]:
    for i, item in enumerate(reversed(items), start=1):
        yield item, items[:-i]


def get_package_map(strings: List[str]) -> Dict[str, List[str]]:
    """Return a dict mapping packages to the modules they contain."""
    m: DefaultDict[str, List[str]] = collections.defaultdict(list)
    for string in strings:
        items = string.split(".")
        for module_name, parents in ancestors(items):
            for _parent in parents:
                package_name = ".".join(parents)
                module_full_name = package_name + "." + module_name
                if module_full_name not in m[package_name]:
                    m[package_name].append(module_full_name)

    return dict(m)


@utils.public
def make_finder(mapping: Dict[str, Dict[str, Any]]) -> PathFinder:
    """Build a Finder that handles modules in ``mapping``."""
    package_map = get_package_map(mapping.keys())
    allowed_names = []
    for lst in package_map.values():
        for sub in lst:
            allowed_names.append(sub)

    allowed_names += list(package_map.keys())
    return PathFinder(allowed_names, mapping)



"""Functions for building a grammar and parsing commands.

Combine rules from multiple modules into a single Lark grammar.
"""

from __future__ import annotations

import re
import textwrap
import types
import unicodedata

from typing import Tuple
from typing import List
from typing import Iterable


import lark


from voca import log
from voca import utils


class Transformer(lark.Transformer):
    def chord(self, arg):
        return arg[0]

    def key(self, arg):

        return arg[0]

    text = list


@log.log_call
def extract(tree: lark.Tree) -> Tuple[str, List]:
    """Transform the parse tree to extract node type and children."""
    tree = Transformer().transform(tree)
    return tree.data, tree.children


@utils.public
@log.log_call
def extract_commands(tree: lark.Tree) -> Tuple[str, List]:
    """Transform the parse tree to extract the node's children."""
    tree = Transformer().transform(tree)

    return tree.children


def _replace_match(match: re.Match) -> str:
    """Replace a symbol with its name."""
    body = unicodedata.name(match.group(0)).replace(" ", "_")
    return f"_{body}_"


@log.log_call
def normalize_pattern(text):
    """Create a human-readable ascii-lowercase form of the rule."""
    return re.sub(r"\W", _replace_match, text).lower()


@utils.public
@log.log_call
def build_rules(registry: utils.Registry) -> List[utils.Rule]:
    """Build a list of rules and attach human-readable names."""
    rules = []
    for i, (pattern, function) in enumerate(registry.pattern_to_function.items()):
        normalized_pattern = normalize_pattern(pattern)
        name = f"rule_{i}__{normalized_pattern}"
        rules.append(utils.Rule(name=name, pattern=pattern, function=function))

    return rules


@utils.public
@log.log_call
def build_grammar(registry: utils.Registry, rules: List[utils.Rule]) -> str:
    """Build a lark grammar string by combining rules and definitions in the registry."""

    start = "?start : message_group"

    message_group = "message_group : message+"

    message = "?message : " + "|".join(rule.name for rule in rules)

    rules_segment = "\n".join([f"{rule.name} : {rule.pattern}" for rule in rules])
    patterns_segment = "\n".join([f"{k} : {v}" for k, v in registry.patterns.items()])
    body = "\n".join([message_group, message, rules_segment, patterns_segment])

    imports = textwrap.dedent(
        """\


        %import common.ESCAPED_STRING
        %import common.SIGNED_NUMBER
        %import common.WS
        %import common.CNAME -> NAME
        %import common.NUMBER
        %import common.WS_INLINE

        %ignore WS

        """
    )
    return "\n".join([start, body, imports])


@utils.public
@log.log_call
def combine_modules(modules: Iterable[utils.PluginModule]) -> utils.WrapperGroup:
    """Combine the wrappers of multiple modules into a single WrapperGroup."""
    wrapper_group = utils.WrapperGroup()
    for module in modules:
        wrapper_group.wrappers.append(module.wrapper)

    return wrapper_group


def combine_registries(registries):
    """Combine several registries together into a single registry."""
    combined = utils.Registry()
    for registry in registries:
        combined.pattern_to_function.update(registry.pattern_to_function)
        combined.patterns.update(registry.patterns)
    return combined



import pytest

from voca import patching

def test_patch():
    mapping = {
        "a.b.c": {"x": 1},
        "a.b.d": {"y": 2},
        "a.b": {"qq": 55},
        "g.h.i": {"x": 44},
        "g.h.j": {"x": 66},
    }

    finder = patching.make_finder(mapping)
    with patching.finder_patch(finder):
        import a.b.c
        import a.b.d
        import collections
        import g.h.i



    assert a.b.qq == 55
    assert a.b.c.x == 1
    assert a.b.d.y == 2
    assert collections.Counter("xyy")["y"] == 2
    assert g.h.i.x == 44

    with pytest.raises(ImportError):
        import g.h.j


def test_get_package_map():
    modules = ["a.b", "a.b.c", "a.b.d", "a.b.c.e", "g.h.i.j"]
    expected = {
        "a": ["a.b"],
        "a.b": ["a.b.c", "a.b.d"],
        "a.b.c": ["a.b.c.e"],
        "g": ["g.h"],
        "g.h": ["g.h.i"],
        "g.h.i": ["g.h.i.j"],
    }
    assert patching.get_package_map(modules) == expected



import functools
import os

import pytest
import trio


from voca import context


async def test_get_current_window_title(turtle_window):

    title = await context.get_current_window_title()
    assert title == "Python Turtle Graphics"



import os
import sys

import eliot

import voca.log


def main(should_log: bool) -> None:
    """
    Set up the app execution environment.

    Args:
        should_log: Enable or disable logging.

    """

    if not should_log:
        return

    eliot.register_exception_extractor(Exception, voca.log.summarize_exception)



import functools
import sys
import time
import signal
import os

from typing import List
from typing import Callable
from typing import Coroutine
from typing import Awaitable


import trio
import eliot

from voca import utils
from voca import platforms
from voca import log


try:
    import pyautogui
except KeyError as e:
    # Display failed
    pyautogui = utils.ModuleLazyRaise("pyautogui", e)

try:
    import pynput
    import pynput.keyboard
except Exception as e:
    pynput = utils.ModuleLazyRaise("pynput", e)


registry = utils.Registry()

v2p = utils.value_to_pronunciation()
registry.define(
    {
        "?any_text": r"/\w.+/",
        "key": utils.regex("|".join(utils.pronunciation_to_value().keys())),
        "chord": 'key ("+" chord)*',
        "?sigstr": utils.regex(
            "|".join(["SIGUSR1", "SIGUSR2"] + list(v2p[str(x)] for x in range(20)))
        ),
        "?pid": r"/\d+/",
    }
)


@log.log_call
def type_chord(chord: utils.KeyChord):
    """Press a key chord. To avoid blocking, call this in a thread."""
    keyboard = pynput.keyboard.Controller()
    modifiers = [getattr(pynput.keyboard.Key, mod.name) for mod in chord.modifiers]

    try:
        key = pynput.keyboard.Key[chord.name]
    except KeyError:
        key = chord.name

    if modifiers:
        with keyboard.pressed(*modifiers):
            # XXX This seems like it wouldn't work, but it is working in tests.
            keyboard.press(key)
            keyboard.release(key)
            pyautogui.press(chord.name)

    else:
        # I'm not sure why pynput wasn't working on simple keys, but this seems to work.
        pyautogui.press(chord.name)


@log.log_async_call
async def press(chord: str):
    """Press a key chord."""
    if isinstance(chord, str):
        await trio.run_sync_in_worker_thread(
            functools.partial(pyautogui.typewrite, [chord])
        )
        return

    await trio.run_sync_in_worker_thread(type_chord, chord)


@log.log_async_call
async def write(message: str):
    """Type the ``message``."""
    await trio.run_sync_in_worker_thread(
        functools.partial(pyautogui.typewrite, message)
    )


@registry.register('"alert" any_text')
@log.log_async_call
async def _alert(text: str):
    """Show a gui alert with ``text``."""
    await trio.run_sync_in_worker_thread(functools.partial(pyautogui.alert, text))


@log.log_async_call
async def speak(message: str):
    """Speak ``message`` aloud."""
    await utils.run_subprocess(["say", message])


@registry.register("chord")
@registry.register('"say" chord')
@log.log_async_call
async def _say(message: List[str]):
    """Press a key."""
    [chord_string] = message
    chord_value = utils.pronunciation_to_value()[chord_string]
    await press(chord_value)


@registry.register('"switch" chord')
@log.log_async_call
async def _switch(message: List[str]):
    """Press super + key."""
    [chord_string] = message
    chord_value = utils.pronunciation_to_value()[chord_string]
    await press(f"super+{chord_value}")


@registry.register('"signal" pid sigstr')
async def send_signal(message):
    """Send a signal to a process."""

    pid, sigstr = message
    try:
        sig = int(sigstr)
    except ValueError:
        sig = getattr(signal, sigstr)

    with eliot.start_action(action_type="kill_signal", signum=sig, pid=pid):
        os.kill(int(pid), sig)


press_key: Callable = utils.async_runner(press)


registry.pattern_to_function['"monitor"'] = press_key("M")
registry.pattern_to_function['"mouse"'] = press_key("O")


@registry.register('"div0"')
async def _div0(*args):
    1 / 0


wrapper = utils.Wrapper(registry)



"""Definition of the worker process.

Each worker is started by the manager. It loads the grammar from the specified
modules matching in the current context, and it executes the received command
before exiting.

The worker sends its logs up to the manager over stdout rather than writing
directly to the log file so the manager can make sure log lines are interleaved
without overlapping in the output.
"""

import importlib
import functools
import sys
import os
import textwrap
import json
import types
import shutil
import pathlib
import importlib.util


from typing import Iterable
from typing import List
from typing import Tuple

import eliot
import trio
import toml
import lark


from voca import utils
from voca import streaming
from voca import log
from voca import parsing
from voca import context
from voca import config


@log.log_async_call
async def handle_message(wrapper_group: utils.WrapperGroup, data: dict):
    """Execute the command in ``data`` with the ``wrapper_group`` containing the grammar."""
    message = data["result"]["hypotheses"][0]["transcript"]

    with eliot.start_action(action_type="parse_command") as action:

        handler = await make_specific_handler(wrapper_group, data)
        tree = handler.parser.parse(message)

    commands = parsing.extract_commands(tree)

    for command in commands:
        rule_name, args = command.data, command.children
        function = handler.rule_name_to_function[rule_name]
        with eliot.start_action(
            action_type="run_command", command=rule_name, args=args, function=function
        ):
            await function(args)


@log.log_call
def load_from_path(import_path: str, filename: str) -> types.ModuleType:
    """Load a module from a filesystem path."""
    spec = importlib.util.spec_from_file_location(import_path, filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


@log.log_call
def get_backup_module(import_path: str, backup_dir: pathlib.Path) -> types.ModuleType:
    """Import a module from the backup directory."""
    sys.path.insert(0, str(backup_dir))
    try:
        module = importlib.import_module(import_path)
    except Exception:
        module = None
    finally:
        del sys.path[0]
    return module


@log.log_call
def save_backup_module(
    module: types.ModuleType, import_path: str, backup_dir: pathlib.Path
) -> None:
    """Save a module in the backup directory."""
    new_path = (backup_dir / import_path.replace(".", "/")).with_suffix(".py")
    new_path.parent.mkdir(exist_ok=True, parents=True)
    shutil.copy2(module.__file__, new_path)


@log.log_call
def get_module(
    import_path: str, backup_dir: pathlib.Path, use_backup_modules: bool
) -> types.ModuleType:
    """Import module and cache it in backup_dir, returning backup on failure."""

    try:
        with eliot.start_action(
            action_type="import_module",
            import_path=import_path,
            sys_path=sys.path,
            sys_meta_path=sys.meta_path,
        ):
            module = importlib.import_module(import_path)
    except Exception:
        if not use_backup_modules:
            raise
        module = get_backup_module(import_path, backup_dir)
    else:
        save_backup_module(module, import_path, backup_dir)
    return module


@log.log_call
def collect_modules(
    import_paths: Iterable[str], use_backup_modules: bool
) -> List[types.ModuleType]:
    """Collect modules from import paths, optionally defaulting to backup modules on failure."""

    backup_dir = pathlib.Path(config.get_config_dir()) / "backup_modules"

    modules = []
    for import_path in import_paths:
        module = get_module(import_path, backup_dir, use_backup_modules)
        if module is not None:
            modules.append(module)
    return modules


def combine_registries(registries: utils.Registry) -> utils.Registry:
    """Combine multiple registries into a single one."""
    combined = utils.Registry()
    for registry in registries:
        combined.pattern_to_function.update(registry.pattern_to_function)
        combined.patterns.update(registry.patterns)
    return combined


async def make_specific_handler(wrapper_group: utils.WrapperGroup, data: dict):
    """Build the command handler for the specific context."""
    filtered = await context.filter_wrappers(wrapper_group, data)
    registry = combine_registries([wrapper.registry for wrapper in filtered.wrappers])

    rules = parsing.build_rules(registry)
    grammar = parsing.build_grammar(registry, rules)
    rule_name_to_function = {rule.name: rule.function for rule in rules}

    parser = lark.Lark(
        grammar, debug=True, lexer="dynamic_complete", maybe_placeholders=True
    )

    return utils.Handler(
        registry=registry, parser=parser, rule_name_to_function=rule_name_to_function
    )


@log.log_async_call
async def async_main(wrapper_group: utils.WrapperGroup):
    """Process input commands as newline-separated json on stdin."""
    stream = trio._unix_pipes.PipeReceiveStream(os.dup(0))
    receiver = streaming.TerminatedFrameReceiver(stream, b"\n")

    async for message_bytes in receiver:
        data = json.loads(message_bytes.decode())

        try:
            with eliot.Action.continue_task(
                task_id=data.get("eliot_task_id", "@")
            ) as action:
                await handle_message(wrapper_group=wrapper_group, data=data)
        except Exception as e:
            action.finish(e)
            raise

        sys.exit(0)


@utils.public
@log.log_call
def main(import_paths: Tuple[str], use_backup_modules: bool):
    """Get the wrapper group and start the event loop."""

    sys.path.insert(0, str(config.get_config_dir()))
    modules = collect_modules(import_paths, use_backup_modules)
    modules = [utils.transform_module(module) for module in modules]

    wrapper_group = parsing.combine_modules(modules)

    trio.run(functools.partial(async_main, wrapper_group=wrapper_group))



import json
import subprocess
import sys
import os
import signal
import time

from tests import helpers


def test_manager_time():
    pid = os.getpid()
    sig = "SIGUSR1"
    num_utterances = 10
    utterances = [f"signal {pid} {sig}"] * num_utterances

    received_at = []

    def handler(_signum, _frame):
        received_at.append(time.perf_counter())

    signal.signal(signal.SIGUSR1, handler)

    rows = [helpers.make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()
    proc = subprocess.Popen(
        [sys.executable, "-m", "voca", "manage", "-i" "voca.plugins.basic"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env={**os.environ, "VOCA_PATCH_CASTER": "1"},
    )

    # Wait for startup.
    time.sleep(10)

    start = time.perf_counter()

    try:
        proc.communicate(input=lines, timeout=10)
    except subprocess.TimeoutExpired:
        pass

    time.sleep(10)
    proc.terminate()

    durations = [when - start for when in received_at]

    assert len(durations) == num_utterances
    assert durations == sorted(durations)

    assert durations[0] < 0.1
    assert durations < [1.0] * num_utterances



"""Define the user configuration.

The location of the configuration directory is determined by platform standards,
pulled from https://pypi.org/project/appdirs/.

"""

import os
import pathlib

import appdirs

from voca import utils


@utils.public
def get_config_dir() -> pathlib.Path:
    """Get the location of voca user config directory."""
    return pathlib.Path(
        os.environ.get("VOCA_CONFIG_DIR") or appdirs.user_config_dir("voca")
    )



# thanks to Casper for contributing commands to this.

from dragonfly import (
    Grammar,
    Context,
    AppContext,
    Dictation,
    Repeat,
    Function,
    Choice,
    Mouse,
    Pause,
)

from castervoice.lib import control
from castervoice.lib import settings
from castervoice.lib.actions import Key, Text
from castervoice.lib.context import AppContext
from castervoice.lib.dfplus.additions import IntegerRefST
from castervoice.lib.dfplus.merge import gfilter
from castervoice.lib.dfplus.merge.mergerule import MergeRule
from castervoice.lib.dfplus.state.short import R
from castervoice.lib.dfplus.merge.ccrmerger import CCRMerger


def findNthToken(text, n, direction):
    Key("c-f").execute()
    Text("%(text)s").execute({"text": text})
    if direction == "reverse":
        # print("yeah? %(n)d")
        Key("s-enter:%(n)d").execute()
    else:
        Key("enter:%(n)d").execute()
        # print("no? %(n)d")
    Key("escape").execute()


class VSCodeCcrRule(MergeRule):
    pronunciation = "visual studio code continuous"
    mwith = CCRMerger.CORE

    mapping = {
        # Note: If you get the bad grammar grammar too complex error, move some of these commands into the non-CCR rule
        # cursor/line navigation
        "simple": R(Key("x")),
        "scroll up [<n>]": R(
            Key("c-up") * Repeat(extra="n"),
            rdescript="VS Code: Scroll Up One Line at a Time",
        ),
        "scroll down [<n>]": R(
            Key("c-down") * Repeat(extra="n"),
            rdescript="VS Code: Scroll Down One Line at a Time",
        ),
        "scroll page up [<n>]": R(
            Key("a-pgup") * Repeat(extra="n"),
            rdescript="VS Code: Scroll Up One Page Up at a Time",
        ),
        "scroll page down [<n>]": R(
            Key("a-pgdown") * Repeat(extra="n"),
            rdescript="VS Code: Scroll Down One Page Down At a Time",
        ),
        "(unindent|outdent) [<n>]": R(
            Key("s-tab") * Repeat(extra="n"), rdescript="VS Code: Unindent"
        ),
        "comment line": R(Key("c-slash"), rdescript="VS Code: Line Comment"),
        "block comment": R(Key("sa-a"), rdescript="VS Code: Block Comment"),
        # Multi-cursor and selection
        "cursor above [<n>]": R(
            Key("ca-up") * Repeat(extra="n"), rdescript="VS Code: Insert Cursor Above"
        ),
        "cursor below [<n>]": R(
            Key("ca-down") * Repeat(extra="n"), rdescript="VS Code: Insert Cursor Above"
        ),
        "remove cursor": R(
            Key("csa-down"), rdescript="VS Code: Remove Cursor"
        ),  # not sure if this command works always; also try csa-up
        # csa-down/up seems to work sometimes to remove 1 of the cursors
        # but I don't really understand how this works
        "tall cursor up": R(
            Key("csa-pgup"), rdescript="VS Code: Add Cursors All The Way Up"
        ),
        "tall cursor down": R(
            Key("csa-pgdown"), rdescript="VS Code: Add Cursors All The Way Down"
        ),
        # Command below requires "brackets select" extension for VS code
        "select [in] brackets [<n>]": R(
            Key("ca-a") * Repeat(extra="n"),
            rdescript="VS Code: Select in between parable punctuation inclusive using 'brackets select' extension",
        )
        * Repeat(extra="n"),
        "all current selection": R(
            Key("c-l"), rdescript="VS Code:Select All Occurrences of Current Selection"
        ),
        "all current word": R(
            Key("c-f2"), rdescript="VS Code: Select All Occurrences of Current Word"
        ),
        "select next [<n>]": R(
            Key("c-f3") * Repeat(extra="n"),
            rdescript="VS Code: Select Next Occurrence of Current Word",
        ),
        "go to next [<n>]": R(
            Key("sa-right/2, c-f3, c-left/2, escape") * Repeat(extra="n"),
            rdescript="VS Code: Go to Next Occurrence of Current Word",
        ),
        # may or may not want the escape afterwards to close the find box
        # note the above command might sometimes be off by one so you have to say one higher
        # than what you mean e.g. if the cursor is at the beginning of the word rather
        # than in the middle or end, you will have to say "next word two" to get to the next word
        "select prior [<n>]": R(
            Key("cs-f3"), rdescript="VS Code: Select Prior Occurrence of Current Word"
        ),
        "go to prior [<n>]": R(
            Key("sa-right/2, cs-f3, c-left/2, escape") * Repeat(extra="n"),
            rdescript="VS Code: Go to Prior Occurrence of Current Word",
        ),
        # may or may not want the escape afterwards to close the find box
        "cursor all": R(
            Key("cs-l"),
            rdescript="VS Code: Add Cursor to All Occurrences of Current Selection",
        ),
        "next cursor [<n>]": R(
            Key("c-d") * Repeat(extra="n"),
            rdescript="VS Code: Add Cursor to Next Occurrence of Current Selection",
        ),
        "indent [<n>]": R(Key("tab") * Repeat(extra="n"), rdescript="VS Code: Indent"),
        "hard delete [<n>]": R(
            Key("s-del"), rdescript="VS Code: Eliminates Line not Just the Text on it"
        ),
        "copy line up [<n>]": R(
            Key("sa-up") * Repeat(extra="n"), rdescript="VS Code: Duplicate Line Above"
        ),
        "copy line up [<n>]": R(
            Key("sa-down") * Repeat(extra="n"),
            rdescript="VS Code: Duplicate Line Below",
        ),
        "switch line down [<n>]": R(
            Key("a-down") * Repeat(extra="n"),
            rdescript="VS Code: Switch Line With the One Below it",
        ),
        "switch line up [<n>]": R(
            Key("a-up") * Repeat(extra="n"),
            rdescript="VS Code: Switch Line With the One Above it",
        ),
        "match bracket": R(
            Key("cs-backslash"), rdescript="VS Code: Jump to Matching Bracket"
        ),
        # commands for selecting between parable characters using "quick and simple text selection" VScode extension (required)
        # repetition of these commands by saying the number expands the selection to include the text between the next (i.e. outer) set of parable characters of the given type
        "select between <between_parables> [<n>]": R(
            Key("c-k, %(between_parables)s") * Repeat(extra="n"),
            rdescript="VS Code: Select between parentheses noninclusive using 'quick and simple text selection' VScode extension",
        ),
        "select around <around_parables> [<n>]": R(
            Key("c-k, %(around_parables)s") * Repeat(extra="n"),
            rdescript="VS Code: Select between parentheses inclusive using 'quick and simple text selection' VScode extension",
        ),
    }
    extras = [
        Dictation("text"),
        Dictation("mim"),
        IntegerRefST("n", 1, 100),
        IntegerRefST("m", 1, 10),
        Choice(
            "between_parables",
            {
                "prekris": "lparen",
                "brax": "lbracket",
                "curly": "lbrace",
                "angle": "langle",
                "single": "squote",
                "quote": "dquote",
            },
        ),
        Choice(
            "around_parables",
            {
                "prekris": "rparen",
                "brax": "rbracket",
                "curly": "rbrace",
                "angle": "rangle",
            },
        ),
    ]

    defaults = {"n": 1, "mim": "", "text": ""}


class VSCodeNonCcrRule(MergeRule):

    pronunciation = "Visual Studio Code Non Continuous"
    mapping = {
        # Moving around a file
        "[(go to | jump | jump to)] line <n>": R(
            Key("c-g") + Text("%(n)d") + Key("enter"), rdescript="VS Code: Go to Line"
        ),
        "go back <n>": R(
            Key("a-left") * Repeat(extra="n"), rdescript="VS Code: Go Back"
        ),
        "go forward [<n>]": R(Key("a-right"), rdescript="VS Code: Go Forward")
        * Repeat(extra="n"),
        # Display
        # note that most of these can be turned on/off with the same command
        "[toggle] full screen": R(
            Key("sa-enter"), rdescript="VS Code: VS Code: Fullscreen"
        ),
        "toggle orientation": R(Key("sa-0"), rdescript="VS Code: Toggle Orientation"),
        "zoom in [<n>]": R(
            Key("c-equal") * Repeat(extra="n"), rdescript="VS Code: Zoom In"
        ),
        "zoom out [<n>]": R(
            Key("c-minus") * Repeat(extra="n"), rdescript="VS Code: Zoom Out"
        ),
        "sidebar": R(Key("c-b"), rdescript="VS Code: Sidebar"),
        "explorer": R(Key("cs-e"), rdescript="VS Code: Explorer"),
        "source control": R(Key("cs-g"), rdescript="VS Code: Source Control"),
        "keyboard shortcuts": R(
            Key("c-k, c-s"), rdescript="VS Code: Keyboard Shortcuts"
        ),
        "key mappings": R(Key("c-k, c-s:2"), rdescript="VS Code: Key Mappings"),
        "settings": R(Key("a-f, p, s"), rdescript="VS Code: User/workspace Settings"),
        "snippets": R(Key("a-f, p, s:2"), rdescript="VS Code: User Snippets"),
        "extensions": R(Key("cs-x"), rdescript="VS Code: Extensions"),
        "search details": R(Key("cs-j"), rdescript="VS Code: Search Details"),
        "output panel": R(Key("cs-u"), rdescript="VS Code: Output Panel"),
        "markdown preview": R(Key("cs-v"), rdescript="VS Code: Markdown Preview"),
        "markdown preview side": R(
            Key("c-k, v"), rdescript="VS Code: Open Markdown Preview to the Side"
        ),
        "Zen mode":
        # note: use esc esc to exit
        R(Key("c-k, z"), rdescript="VS Code: Zen mode"),
        # File Management
        "[open] command palette": R(Key("cs-p"), rdescript="VS Code: Command Palette"),
        "(open [file] | go to [tab]) [<text>]": R(
            Key("c-p") + Text("%(text)s"), rdescript="VS Code: Go to File"
        ),
        "Save and close": R(
            Key("c-s/10, c-w"), rdescript="VS Code: Save And Close File"
        ),
        "new file": R(Key("c-n"), rdescript="VS Code: New File"),
        "new window": R(Key("cs-n"), rdescript="VS Code: New Window"),
        "close window": R(Key("a-f4"), rdescript="VS Code: Close Window"),
        "close workspace": R(Key("c-k, f"), rdescript="VS Code: Close Workspace"),
        "close editor": R(Key("c-f4"), rdescript="VS Code: Close Editor"),
        "save as": R(Key("cs-s"), rdescript="VS Code: Save As"),
        "save all": R(Key("c-k, s"), rdescript="VS Code: Save all"),
        "next tab [<n>]": R(
            Key("c-pgdown") * Repeat(extra="n"), rdescript="VS Code: Next Tab"
        ),
        "previous tab [<n>]": R(
            Key("c-pgup") * Repeat(extra="n"), rdescript="VS Code: Previous Tab"
        ),
        "close tab [<n>]": R(
            Key("c-f4/20") * Repeat(extra="n"),
            rdescript="VS Code: Visual Studio: Close Tab",
        ),
        "(recent | R) tab [<n>]": R(
            Key("c-tab") * Repeat(extra="n"), rdescript="VS Code: Go to Most Recent Tab"
        ),
        "reopen tab [<n>]": R(
            Key("cs-t") * Repeat(extra="n"),
            rdescript="VS Code: Reopen Most Recently Tab",
        ),
        "Exit preview": R(Key("space, c-z"), rdescript="VS Code: Exit Preview"),
        "keep preview open": R(
            Key("c-k, enter"), rdescript="VS Code: Keep Preview Mode Editor Open"
        ),
        "copy path": R(Key("c-k, p"), rdescript="VS Code: Copy Path of Active File"),
        "windows explorer here": R(
            Key("c-k, r"),
            rdescript="VS Code: open Windows Explorer to the Location of Active File",
        ),
        "show active file in new window": R(
            Key("c-k, o"), rdescript="VS Code: Show Active File in New Window"
        ),
        # Search
        "(search | find)": R(Key("c-f"), rdescript="VS Code: Find in File"),
        "replace": R(Key("c-h"), rdescript="VS Code: Replace"),
        "find in files": R(Key("cs-f"), rdescript="VS Code: Find in Files"),
        "replace in files": R(Key("cs-h"), rdescript="VS Code: Replace in Files"),
        "next find": R(Key("f3"), rdescript="VS Code: Go to Next Occurrence"),
        "(prior | previous) find": R(
            Key("s-f3"), rdescript="VS Code: Go to Previous Occurrence"
        ),
        "select all occurrences": R(
            Key("a-enter"), rdescript="VS Code: Select all Occurrences of Find Match"
        ),
        "toggle case sensitive": R(
            Key("a-c"), rdescript="VS Code: Toggle Case Sensitive"
        ),
        "toggle regex": R(Key("a-r"), rdescript="VS Code: Toggle Regular Expressions"),
        "toggle whole word": R(Key("a-w"), rdescript="VS Code: Toggle Whole Word"),
        "(find | jump [to]) next <text>": R(
            Function(findNthToken, n=1, direction="forward"),
            rdescript="VS Code: Find Next",
        ),
        "(find | jump [to]) previous <text>": R(
            Function(findNthToken, n=1, direction="reverse"),
            rdescript="VS Code: Find Previous",
        ),
        "show all symbols": R(Key("c-t"), rdescript="VS Code: Show all Symbols"),
        "go to symbol": R(Key("cs-o"), rdescript="VS Code: Go to Symbol"),
        # Editor Management
        "close editor": R(Key("c-w"), rdescript="VS Code: Close Editor"),
        "close folder": R(Key("c-k, f"), rdescript="VS Code: Close Folder"),
        "split editor": R(
            Key("c-backslash"), rdescript="VS Code: Split Editor into 2 Panes"
        ),
        "next pane": R(Key("c-k, c-right"), rdescript="VS Code: Move to Next Pane"),
        "(prior | previous | un) pane": R(
            Key("c-k, c-right"), rdescript="VS Code: Move to Next Pane"
        ),
        "shift group left": R(
            Key("c-k, left"),
            rdescript="Shift Current Group of Tabs to the Left E.g. Swap with Pane to the Left",
        ),
        "shift group left": R(
            Key("c-k, right"),
            rdescript="Shift Current Group of Tabs to the Right E.g. Swap with Pane to the Right",
        ),
        "<nth> tab": R(Key("c-%(nth)s"), rdescript="VS Code: Go to Nth Pane"),
        # Languages Editing
        "go to definition": R(Key("f12"), rdescript="VS Code: Go to Definition"),
        "go to required definition": R(
            Key("c-f12:2, c-right:5, left/50, f12"),
            rdescript="VS Code: Go to Required Definition",
        ),
        "peak definition": R(Key("a-f12"), rdescript="VS Code: Peak Definition"),
        "trigger parameter hints": R(
            Key("cs-space"), rdescript="VS Code: Trigger Parameter Hints"
        ),
        "format that": R(Key("c-k, c-f"), rdescript="VS Code: Format Selection"),
        "(definition to side | side def)": R(
            Key("c-k, f12"), rdescript="VS Code: Open Definition to the Side"
        ),
        "show references": R(Key("s-f12"), rdescript="VS Code: Show References"),
        "rename symbol": R(Key("f2"), rdescript="VS Code: Rename Symbol"),
        "(trim white)": R(
            Key("c-k, c-x"), rdescript="VS Code: Trim Trailing White Space"
        ),
        "change file language": R(
            Key("c-k, m"), rdescript="VS Code: Change File Language"
        ),
        # Debugging
        "debug": R(Key("cs-d"), rdescript="VS Code: Debug"),
        "[toggle] breakpoint": R(Key("f9"), rdescript="VS Code: Breakpoint"),
        "step over [<n>]": R(
            Key("f10/50") * Repeat(extra="n"), rdescript="VS Code: Step Over"
        ),
        "step into": R(Key("f11"), rdescript="VS Code: Step into"),
        "step out [of]": R(Key("s-f11"), rdescript="VS Code: Step Out"),
        "resume": R(Key("f5"), rdescript="VS Code: Resume"),
        "stopper": R(Key("s-f5"), rdescript="VS Code: Debug Stop"),
        "continue": R(Key("f5"), rdescript="VS Code: Start/Continue"),
        "show hover": R(
            Key("c-k, c-i"),
            rdescript="Show the little box as if you are hovering your mouse over the place where the cursor (As opposed to the mouse pointer) currently is",
        ),
        "[show] problems [panel]": R(
            Key("cs-m"), rdescript="VS Code: Show Problems Panel"
        ),
        "next error": R(
            Key("f8"), rdescript="VS Code: Go to Next Error"
        ),  # doesn't seem to be working properly
        "(prior | previous) error": R(
            Key("s-f8"), rdescript="VS Code: Go to Previous Error"
        ),
        "toggle tab moves focus": R(
            Key("c-m"), rdescript="VS Code: Toggle Taboos Focus"
        ),
        # Integrated Terminal
        "[show] terminal": R(
            Key("c-backtick"), rdescript="VS Code: Show Integrated Terminal"
        ),
        "new terminal": R(
            Key("cs-backtick"), rdescript="VS Code: New Integrated Terminal"
        ),
        "terminal scroll up": R(Key("c-up"), rdescript="VS Code: Terminal Scroll Up"),
        "terminal scroll down": R(
            Key("c-down"), rdescript="VS Code: Terminal Scroll Down"
        ),
        "terminal page up": R(Key("s-pgup"), rdescript="VS Code: Terminal Page Up"),
        "terminal page down": R(
            Key("s-pgdown"), rdescript="VS Code: Terminal Page Down"
        ),
        # Collapsing
        "(fold | collapse) region": R(
            Key("cs-lbracket"), rdescript="VS Code: Collapse Region"
        ),
        "(unfold | uncollapse) region": R(
            Key("cs-rbracket"), rdescript="VS Code: Uncollapse Region"
        ),
        "(fold | collapse) [all] subregions": R(
            Key("c-k, c-lbracket"), rdescript="VS Code: Collapse all Subregions"
        ),
        "(unfold | uncollapse) [all] subregions": R(
            Key("c-k, c-rbracket"), rdescript="VS Code: Uncollapse all Subregions"
        ),
        "(fold | collapse) [all] regions": R(
            Key("c-k, c-0"), rdescript="VS Code: Safe Collapse all Regions"
        ),
        "(unfold | uncollapse) [all] regions": R(
            Key("c-k, c-j"), rdescript="VS Code: On Collapse all Regions"
        ),
        "toggle word wrap": R(Key("a-z"), rdescript="VS Code: Toggle Word Wrap"),
        "run this line": R(Key("csa-l"), rdescript="VS Code: Run This Line"),
        "join line": R(Key("csa-j"), rdescript="VS Code: Join Line"),
        # requires gitlens extension
        "toggle blame": R(Key("cs-g, b"), rdescript="VS Code: Toggle Blame"),
        "lens commit details": R(
            Key("cs-g, c"), rdescript="VS Code: Lens Commit Details"
        ),
        "lens file history": R(Key("cs-g, h"), rdescript="VS Code: Lens File History"),
        "lens repo status": R(Key("cs-g, s"), rdescript="VS Code: Lens Repo Status"),
        "toggle git lens": R(Key("cs-g, s-b"), rdescript="VS Code: Toggle Git Lens"),
        # requires bookmark extension
        "mark (prev | prior | previous)": R(
            Key("ca-j"), rdescript="VS Code: Mark Previous"
        ),
        "mark next": R(Key("ca-l"), rdescript="VS Code: Mark Next"),
    }
    extras = [
        Dictation("text"),
        Dictation("mim"),
        IntegerRefST("n", 1, 1000),
        Choice(
            "nth",
            {
                "first": "1",
                "second": "2",
                "third": "3",
                "fourth": "4",
                "fifth": "5",
                "sixth": "6",
            },
        ),
    ]
    defaults = {"n": 1, "mim": "", "text": ""}


# ---------------------------------------------------------------------------

# initialise the rule.
context = AppContext(title="Visual Studio Code", executable="code")
grammar = Grammar("Visual Studio Code", context=context)
if settings.SETTINGS["apps"]["visualstudiocode"]:
    if settings.SETTINGS["miscellaneous"]["rdp_mode"]:
        control.nexus().merger.add_global_rule(VSCodeCcrRule())
    else:
        control.nexus().merger.add_app_rule(VSCodeCcrRule(), context)



"""Provide logging functions.

Eliot is used for structured logging.
"""

from __future__ import annotations

import collections
import contextlib
import datetime
import functools
import json
import sys
import inspect
import os
import io
import traceback
import types
import pathlib

from typing import Optional
from typing import Callable
from typing import Any
from typing import Union
from typing import Iterable
from typing import Container
from typing import List
from typing import Dict

import attr
import eliot
import six

from voca import config
from voca import utils


@utils.public
@functools.singledispatch
def to_serializable(obj: Any) -> Union[str, list, dict, int, float]:
    try:
        return {"type": type(obj).__name__, "attributes": attr.asdict(obj)}
    except attr.exceptions.NotAnAttrsClassError:
        return str(obj)


@to_serializable.register(types.SimpleNamespace)
def _(obj):
    return vars(obj)


@utils.public
def json_to_file(file: Optional[io.TextIOWrapper] = None) -> Callable:
    """Serialize to json and print into a file, defaulting to stdout."""

    if file is None:

        file = sys.stdout

    def _json_to_file(x):

        print(json.dumps(x, default=to_serializable), file=file)

    return _json_to_file


def _exception_lines(exc: BaseException) -> List[str]:
    """Get a list of traceback string lines from an exception."""
    return traceback.format_exception(type(exc), exc, exc.__traceback__)


def _extract_traceback(exc: BaseException) -> str:
    """Make a traceback string from an exception."""
    return "".join(_exception_lines(exc))


def _exception_data(exc: BaseException) -> Dict[str, Any]:
    """Collect interesting data to report about the exception."""
    exclude = set(dir(Exception())) - {"args", "__cause__", "__context__"}
    return {k: v for k, v in inspect.getmembers(exc) if k not in exclude}


def summarize_exception(exc: BaseException) -> Dict[str, dict]:
    """Combine exception data with traceback into a dict for logging."""
    return {
        "traceback": _extract_traceback(exc),
        "exception_data": _exception_data(exc),
    }


@utils.public
def log_call(
    wrapped_function: Optional[Callable] = None,
    action_type: Optional[str] = None,
    include_args: Optional[Iterable[str]] = None,
    include_result: bool = True,
) -> Callable:
    """Decorator/decorator factory that logs inputs and the return result.
    If used with inputs (i.e. as a decorator factory), it accepts the following
    parameters:
    @param action_type: The action type to use.  If not given the function name
        will be used.
    @param include_args: If given, should be a list of strings, the arguments to log.
    @param include_result: True by default. If False, the return result isn't logged.
    """
    if wrapped_function is None:
        return functools.partial(
            log_call,
            action_type=action_type,
            include_args=include_args,
            include_result=include_result,
        )

    if action_type is None:
        action_type = "{}.{}".format(
            wrapped_function.__module__, wrapped_function.__qualname__
        )

    if include_args is not None:

        sig = inspect.signature(wrapped_function)
        if set(include_args) - set(sig.parameters):
            raise ValueError(
                (
                    "include_args ({}) lists arguments not in the " "wrapped function"
                ).format(include_args)
            )

    @functools.wraps(wrapped_function)
    def logging_wrapper(*args, **kwargs):
        bound_args = inspect.signature(wrapped_function).bind(*args, **kwargs)

        # Remove self if it's included:
        if "self" in bound_args.arguments:
            bound_args.arguments.pop("self")

        # Filter arguments to log, if necessary:
        if include_args is not None:
            bound_args.arguments = {k: bound_args.arguments[k] for k in include_args}

        with eliot.start_action(
            action_type=action_type, **bound_args.arguments
        ) as action:

            result = wrapped_function(*args, **kwargs)
            if not include_result:
                return result
            action.add_success_fields(result=result)
            return result

    return logging_wrapper


@utils.public
def log_async_call(
    wrapped_function: Optional[Callable] = None,
    action_type: Optional[str] = None,
    include_args: Optional[Iterable[str]] = None,
    include_result: bool = True,
) -> Callable:
    """Decorator/decorator factory that logs inputs and the return result.
    If used with inputs (i.e. as a decorator factory), it accepts the following
    parameters:
    @param action_type: The action type to use.  If not given the function name
        will be used.
    @param include_args: If given, should be a list of strings, the arguments to log.
    @param include_result: True by default. If False, the return result isn't logged.
    """
    if wrapped_function is None:
        return functools.partial(
            log_call,
            action_type=action_type,
            include_args=include_args,
            include_result=include_result,
        )

    if action_type is None:

        action_type = "{}.{}".format(
            wrapped_function.__module__, wrapped_function.__qualname__
        )

    if include_args is not None:

        sig = inspect.signature(wrapped_function)
        if set(include_args) - set(sig.parameters):
            raise ValueError(
                f"include_args ({include_args}) lists arguments not in the "
                "wrapped function"
            )

    @functools.wraps(wrapped_function)
    async def logging_wrapper(*args, **kwargs):
        bound_args = inspect.signature(wrapped_function).bind(*args, **kwargs)

        # Remove self if it's included:
        if "self" in bound_args.arguments:
            bound_args.arguments.pop("self")

        # Filter arguments to log, if necessary:
        if include_args is not None:
            bound_args.arguments = {k: bound_args.arguments[k] for k in include_args}

        with eliot.start_action(
            action_type=action_type, **bound_args.arguments
        ) as action:

            result = await wrapped_function(*args, **kwargs)

            if include_result:
                action.add_success_fields(result=result)

            return result

    return logging_wrapper


def get_log_filename() -> pathlib.Path:
    """Return the filename of the session log, creating any parent directories."""

    log_dir = config.get_config_dir() / "logs"
    log_dir.mkdir(exist_ok=True, parents=True)

    timestamp = datetime.datetime.now().isoformat()
    name = timestamp + ".voca-log.jsonl"

    return log_dir / name



#!/usr/bin/env python3
"""Plot the live microphone signal(s) with matplotlib.

Matplotlib and NumPy have to be installed.


This module is from the sounddevice docs.
"""


import argparse
import queue
import sys


def int_or_str(text):
    """Helper function for argument parsing."""
    try:
        return int(text)
    except ValueError:
        return text


parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-l",
    "--list-devices",
    action="store_true",
    help="show list of audio devices and exit",
)
parser.add_argument(
    "-d", "--device", type=int_or_str, help="input device (numeric ID or substring)"
)
parser.add_argument(
    "-w",
    "--window",
    type=float,
    default=200,
    metavar="DURATION",
    help="visible time slot (default: %(default)s ms)",
)
parser.add_argument(
    "-i",
    "--interval",
    type=float,
    default=30,
    help="minimum time between plot updates (default: %(default)s ms)",
)
parser.add_argument("-b", "--blocksize", type=int, help="block size (in samples)")
parser.add_argument(
    "-r", "--samplerate", type=float, help="sampling rate of audio device"
)
parser.add_argument(
    "-n",
    "--downsample",
    type=int,
    default=10,
    metavar="N",
    help="display every Nth sample (default: %(default)s)",
)
parser.add_argument(
    "channels",
    type=int,
    default=[1],
    nargs="*",
    metavar="CHANNEL",
    help="input channels to plot (default: the first)",
)


def audio_callback(indata, frames, time, status):
    """This is called (from a separate thread) for each audio block."""
    if status:
        print(status, file=sys.stderr)
    # Fancy indexing with mapping creates a (necessary!) copy:
    q.put(indata[:: args.downsample, mapping])


def update_plot(frame):
    """This is called by matplotlib for each plot update.

    Typically, audio callbacks happen more frequently than plot updates,
    therefore the queue tends to contain multiple blocks of audio data.

    """
    global plotdata
    while True:
        try:
            data = q.get_nowait()
        except queue.Empty:
            break
        shift = len(data)
        plotdata = np.roll(plotdata, -shift, axis=0)
        plotdata[-shift:, :] = data
    for column, line in enumerate(lines):
        line.set_ydata(plotdata[:, column])
    return lines


if __name__ == "__main__":
    args = parser.parse_args()

    if any(c < 1 for c in args.channels):
        parser.error("argument CHANNEL: must be >= 1")
    mapping = [c - 1 for c in args.channels]  # Channel numbers start with 1
    q = queue.Queue()

    try:
        from matplotlib.animation import FuncAnimation
        import matplotlib.pyplot as plt
        import numpy as np
        import sounddevice as sd

        if args.list_devices:
            print(sd.query_devices())
            parser.exit(0)
        if args.samplerate is None:
            device_info = sd.query_devices(args.device, "input")
            args.samplerate = device_info["default_samplerate"]

        length = int(args.window * args.samplerate / (1000 * args.downsample))
        plotdata = np.zeros((length, len(args.channels)))

        fig, ax = plt.subplots()
        lines = ax.plot(plotdata)
        if len(args.channels) > 1:
            ax.legend(
                ["channel {}".format(c) for c in args.channels],
                loc="lower left",
                ncol=len(args.channels),
            )
        ax.axis((0, len(plotdata), -1, 1))
        ax.set_yticks([0])
        ax.yaxis.grid(True)
        ax.tick_params(
            bottom=False,
            top=False,
            labelbottom=False,
            right=False,
            left=False,
            labelleft=False,
        )
        fig.tight_layout(pad=0)

        stream = sd.InputStream(
            device=args.device,
            channels=max(args.channels),
            samplerate=args.samplerate,
            callback=audio_callback,
        )
        ani = FuncAnimation(fig, update_plot, interval=args.interval, blit=True)
        with stream:
            plt.show()
    except Exception as e:
        parser.exit(type(e).__name__ + ": " + str(e))



"""
Functions for implementing cross-platform functionality.
"""

import enum
import platform

from typing import Tuple
from typing import Callable

from voca import utils

_name_to_function = {}


@utils.public
class System(enum.Enum):
    LINUX = enum.auto()
    WINDOWS = enum.auto()
    DARWIN = enum.auto()


@utils.public
def implementation(*which: Tuple[System]) -> Callable:
    """Decorator for functions that provide a platform-specific functionality.

    To define a function ``create_new_window`` for multiple platforms, use the
    decorator:

    ::

        @implementation(System.LINUX)
        def create_new_window(name):
            return window(name)


        @implementation(System.WINDOWS)
        def create_new_window(name):
            return application(name)


    Then just call it with ``create_new_window()`` and the proper platform
    implementation will be called. Access the dictionary mapping platform to
    implementation function as ``create_new_window.implementations``.

    """

    def wrap(func: Callable) -> Callable:

        try:
            seen_function = _name_to_function[func.__name__]
        except KeyError:
            seen_function = func
            _name_to_function[func.__name__] = func

        func.implementations = getattr(seen_function, "implementations", {})

        for item in which:
            seen_function.implementations[item] = func

        system_enum_value = System[(platform.system().upper())]

        try:
            return seen_function.implementations[system_enum_value]
        except KeyError:
            return seen_function

    return wrap



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import os
import io
import re
import sys
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
import pathlib


from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


PROJECT_ROOT = pathlib.Path(__file__).parent


try:
    with open(PROJECT_ROOT / "requirements.in") as f:
        INSTALL_REQUIRES = [line for line in f.read().splitlines() if line[0].isalpha()]
except FileNotFoundError:
    print(sys.exc_info())
    INSTALL_REQUIRES = []

if os.environ.get("READTHEDOCS") == "True":
    INSTALL_REQUIRES = []

NAME = "voca"

setup(
    name=NAME,
    version="0.1.10",
    description="Control your computer by voice!",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    long_description_content_type="text/x-rst",
    author="Full Name",
    author_email="email@example.com",
    url="https://github.com/python-voca/voca",
    packages=list(find_packages("src")),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        "Topic :: Utilities",
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES
    # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ,
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        "console_scripts": ["voca = voca.cli:cli"],
        "voca_plugins": [
            "basic = voca.plugins:basic",
            "math = voca.plugins:math",
            "python = voca.plugins:python",
            "yes = voca.plugins:yes",
            "no = voca.plugins:no",
            "turtle_context = voca.plugins:turtle_context",
            "vscode = voca.plugins:vscode",
        ],
    },
)



"""Utilities for working with Voca."""


import re
import functools
import subprocess
import types
import importlib


from typing import Any
from typing import List
from typing import Dict
from typing import Optional
from typing import Callable
from typing import MutableMapping
from typing import Mapping
from typing import Awaitable

from typing_extensions import Protocol

import attr
import toml
import trio
import lark
import pkg_resources
import importlib_resources
import public as atpublic

import voca


public = atpublic.public

# XXX There are too many things in here -- it should be split up.


@public
@attr.s
class Registry:
    pattern_to_function: MutableMapping[str, Callable] = attr.ib(factory=dict)
    patterns: MutableMapping = attr.ib(factory=dict)

    def register(self, pattern: str) -> Callable:
        """Decorator registering a pattern to map to a function."""

        def _register(function: Callable) -> Callable:
            self.pattern_to_function[pattern] = function
            return function

        return _register

    def define(self, patterns: Optional[Mapping[str, str]] = None, **kwargs):
        """Define a rule to map to a combination of other known rules."""
        if patterns:
            self.patterns.update(patterns)
        self.patterns.update(kwargs)


@public
def pronunciation_to_value() -> MutableMapping[str, Any]:
    """Return a dict of pronunciation to value."""
    text = importlib_resources.read_text(voca, "pronunciation.toml")
    return toml.loads(text)


@public
def value_to_pronunciation() -> Dict[Any, str]:
    """Return a dict of value to pronunciation."""
    return {v: k for k, v in pronunciation_to_value().items()}


@public
async def run_subprocess(
    command: List[str], *, input=None, capture_output=False, **options: Dict[str, Any]
):
    """Run a subprocess an wait for it to exit."""
    if input is not None:
        options["stdin"] = subprocess.PIPE
    if capture_output:
        options["stdout"] = options["stderr"] = subprocess.PIPE

    stdout_chunks: List = []
    stderr_chunks: List = []

    async with trio.Process(command, **options) as proc:

        async def feed_input():
            async with proc.stdin:
                if input:
                    try:
                        await proc.stdin.send_all(input)
                    except trio.BrokenResourceError:
                        pass

        async def read_output(stream, chunks):
            async with stream:
                while True:
                    chunk = await stream.receive_some(32768)
                    if not chunk:
                        break
                    chunks.append(chunk)

        async with trio.open_nursery() as nursery:
            if proc.stdin is not None:
                nursery.start_soon(feed_input)
            if proc.stdout is not None:
                nursery.start_soon(read_output, proc.stdout, stdout_chunks)
            if proc.stderr is not None:
                nursery.start_soon(read_output, proc.stderr, stderr_chunks)
            await proc.wait()

    stdout = b"".join(stdout_chunks) if proc.stdout is not None else None
    stderr = b"".join(stderr_chunks) if proc.stderr is not None else None

    if proc.returncode:
        raise subprocess.CalledProcessError(
            proc.returncode, proc.args, output=stdout, stderr=stderr
        )
    return subprocess.CompletedProcess(proc.args, proc.returncode, stdout, stderr)


@public
def quote(word: str) -> str:
    """Format a word in single or double quotation marks for lark."""
    if word.startswith('"'):
        return f"'{word}'"
    return f'"{word}"'


@public
def regex(word: str) -> str:
    """Format a word in slashes for lark."""
    return f"/{word}/"


@attr.dataclass
class Handler:
    registry: Registry
    parser: lark.Lark
    rule_name_to_function: dict


@attr.dataclass
class HandlerGroup:
    handlers: List[Handler]

    async def pick_handler(self, data: dict) -> Handler:
        """Select the first handler."""
        return self.handlers[0]


class Context(Protocol):
    async def check(self, data=None) -> bool:
        ...


@attr.dataclass
class AlwaysContext:
    # TODO Move this into context.py.
    async def check(self, data=None) -> bool:
        """Always True."""
        return True


@attr.dataclass
class NeverContext:
    # TODO Move this into context.py.
    async def check(self, data=None) -> bool:
        """Always False."""
        return False


@public
@attr.dataclass
class Wrapper:
    registry: Registry
    context: Context = attr.ib(default=AlwaysContext)


@public
@attr.s
class WrapperGroup:
    wrappers: List[Wrapper] = attr.ib(factory=list)


@attr.dataclass
class Rule:
    name: str
    pattern: str
    function: Callable


@attr.dataclass
class PluginModule(Protocol):
    wrapper: Wrapper


@public
@attr.dataclass
class KeyModifier:
    name: str


@public
@attr.dataclass
class SimpleKey:
    name: str


@public
@attr.dataclass
class KeyChord:
    name: str = attr.ib(default=None)
    modifiers: List[KeyModifier] = attr.ib(factory=list)


@public
def async_runner(async_function: Callable):
    def build(*args, **kwargs) -> Callable:
        @functools.wraps(async_function)
        async def run(_message: str):
            await async_function(*args, **kwargs)

        return run

    return build


@public
def replace(message: str) -> str:
    """Replace pronunciation with its value."""
    lookup = pronunciation_to_value()
    pattern = re.compile(r"\b(" + "|".join(lookup.keys()) + r")\b")
    return pattern.sub(lambda x: lookup[x.group()], message)


@public
def plugin_module_paths() -> List[str]:
    """Get the import paths of the plugin modules."""
    return [
        entry_point.module_name + "." + entry_point.name
        for entry_point in pkg_resources.iter_entry_points("voca_plugins")
    ]


def get_module_names() -> List[str]:
    config_files = config.get_config_dir() / "user_modules".glob("*.py")
    config_modules = [
        "user_modules" + file.with_suffix("").name for file in config_files
    ]
    return config_modules + plugin_module_paths()


MODULE_TRANSFORMERS = []


@public
def transform_module(module: types.ModuleType) -> types.ModuleType:
    """Call each function in MODULE_TRANSFORMERS on the module."""
    for transform in MODULE_TRANSFORMERS:
        module = transform(module)

    return module


@public
@attr.dataclass
class ModuleLazyRaise:
    owner_name: str
    exc: Exception

    def __getattr__(self, name: str):
        raise ImportError(
            "Could not access {name} on {self.owner_name} because {self.owner_name} failed to import, with exception:"
        ) from self.exc



import sys

from typing import List

import sympy
import sympy.parsing.sympy_parser

from voca import utils
from voca.plugins import basic

registry = utils.Registry()
registry.define({"?math": r"/[A-Za-z0-9\-\+ ]+/"})


@registry.register('"solve" math')
@registry.register('"arch" math')
async def _solve(items: List[str]):

    [message] = items
    message = utils.replace(message)
    expr = sympy.parsing.sympy_parser.parse_expr(message)
    result = str(expr.simplify())

    await basic.announce([result])
    await basic.speak(str(result))


wrapper = utils.Wrapper(registry)



import contextlib
import subprocess
import sys
import time

import trio
import sneakysnek.recorder


def run(args, **kwargs):
    return subprocess.check_output(
        [sys.executable, "-m", "voca"] + args, **kwargs
    )


@contextlib.contextmanager
def capture_keypresses():
    captured = []

    def capture(event):
        if event.event == sneakysnek.keyboard_event.KeyboardEvents.DOWN:
            captured.append(event.keyboard_key.name)

    rec = sneakysnek.recorder.Recorder.record(capture)
    yield captured
    time.sleep(2)
    rec.stop()

@contextlib.asynccontextmanager
async def async_capture_keypresses():
    captured = []

    def capture(event):
        if event.event == sneakysnek.keyboard_event.KeyboardEvents.DOWN:
            captured.append(event.keyboard_key.name)

    rec = await trio.run_sync_in_worker_thread(sneakysnek.recorder.Recorder.record, capture)
    yield captured
    rec.stop()



def make_command(utterance, final=True):
    return {
        "status": 0,
        "segment": 0,
        "result": {"hypotheses": [{"transcript": utterance}], "final": final},
        "id": "eec37b79-f55e-4bf8-9afe-01f278902599",
    }
